from builtins import input
from builtins import object
import subprocess
import sys
from .logger import create_logger

logger = create_logger()

class GitCommandRunner(object):

    @staticmethod
    def remove_api_token(config_key):
        return run_cmd("git config --unset %s" % config_key).strip()

    @staticmethod
    def get_api_token(config_key):
        return run_cmd("git config %s" % config_key).strip()

    @staticmethod
    def add_api_token(config_key, val):
        return run_cmd("git config %s %s" % (config_key, val))

    @staticmethod
    def get_last_commit_msg():
        return run_cmd("git log -1 --pretty=%B").strip()

    @staticmethod
    def inside_working_tree():
        in_git_tree = run_cmd("git rev-parse --is-inside-work-tree").strip()
        if in_git_tree == 'true':
            return True
        return False

    @staticmethod
    def get_current_branch():
        branch = run_cmd("git rev-parse --abbrev-ref HEAD").rstrip()
        if 'fatal: Not a git repository' in branch:
            logger.fatal("Unable to determine current git branch: git rev-parse --abbrev-ref HEAD\n")
            sys.exit()

        return branch

    @staticmethod
    def get_remote_origin_url():
        return run_cmd("git config --get remote.origin.url").rstrip()

    @staticmethod
    def push_branch_to_remote(source):
        cmd = "git push origin HEAD:%s" % source
        u = input("\n\nPushing local branch [%s] to remote [%s],  Yes or no? " % (source, source))
        if u.lower() != 'yes' and u.lower() != 'y':
            logger.warn("Exiting...\n")
            sys.exit()

        output = run_cmd(cmd)
        logger.warn(output)

def run_cmd(cmd):
    cmd_list = cmd.split()
    try:
        p = subprocess.Popen(
        cmd_list,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
        output = p.communicate()

        # decode() required for python3 compatability
        if not output[0]:
            return output[1].decode()
        return output[0].decode()

    except OSError as e:
        logger.fatal("'%s' is either not in your path or is not installed\n" % cmd_list[0])
        sys.exit()
    except IndexError as e:
        logger.fatal("Unexpected response from `%s`\n" % cmd)
        sys.exit()


from __future__ import absolute_import
from builtins import object
from .GitHubService import GitHubService
from .GitLabService import GitLabService
from .GitCommandRunner import GitCommandRunner as Git
from .GitUtils import determine_git_domain, repo_url_attributes
from .configprocesser import get_configuration
from .logger import create_logger

import sys

#required depedendies
import requests


class GitServiceManager(object):
    GIT_SERVICES = [GitHubService, GitLabService]

    def __init__(self, target_branch):
        self.logger = create_logger()
        self.git_config_token_key = get_configuration('git_config_token_key')
        self.remote_origin_url = Git.get_remote_origin_url()
        self.source_branch = Git.get_current_branch()
        self.parent_branch = target_branch
        git_domain_url = determine_git_domain(self.remote_origin_url)
        (self.namespace, self.project) = repo_url_attributes(git_domain_url, self.remote_origin_url)

        if not Git.inside_working_tree():
            self.logger.error("you are not in a working git directory. please navigate to your current project")
            sys.exit()

        if self.parent_branch == self.source_branch:
            self.logger.error("source branch (%s) cannot be the same as the one to merge with (%s)" % (self.source_branch, self.parent_branch))
            sys.exit()

        if not git_domain_url:
            self.logger.fatal("Unable to determine git hosting service: %s" % self.remote_origin_url)
            self.logger.fatal("Are you using a relative URL path instead of a (sub)domain?")
            sys.exit()

        if not self.namespace or not self.project:
            self.logger.fatal("Unable to determine project and/or namespace from: %s" % self.remote_origin_url)
            sys.exit()

        self.git_service_engine = self._choose_engine(git_domain_url)
        self._get_set_api_token()


    def _get_set_api_token(self):
        #Get API token
        val = Git.get_api_token(self.git_config_token_key)
        if (not val) or (val == "None"):
            val, err = self.git_service_engine._setup_token()

            if (val== -1):
                if (err):
                    self.logger.fatal(err)
                else:
                    self.logger.fatal('Could not request token. Please try again')
                sys.exit()

            Git.add_api_token(self.git_config_token_key, val)

        return val


    def _choose_engine(self, domain):
        for engine in GitServiceManager.GIT_SERVICES:
            if engine.SERVICE_NAME in domain:
                return engine(self.source_branch, self.parent_branch, domain, self.namespace, self.project, self.logger)
        self.logger.fatal('%s is not a supported service [%s]' % (domain, GIT_SERVICES))


    def post_review(self):
        Git.push_branch_to_remote(self.source_branch)
        token = self._get_set_api_token()

        if not self.git_service_engine.parent_branch_exists(token):
            self.logger.fatal("Target branch does not exist. Please try again")
            sys.exit()

        params = {'message': Git.get_last_commit_msg(), 'api_token': token}

        (msg, error) = self.git_service_engine.issue_pull_request(params)
        if error == 401:
            Git.remove_api_token(self.git_config_token_key)
            self.logger.fatal("API token is invalid. Please re-run the command to create a new token")
            sys.exit()

        self.logger.warn(msg+'\n')



from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import input
from .BaseService import BaseService

import sys
import json
import re
import random

#Dependencies
from urllib.parse import urljoin
import requests
import getpass

class GitHubService(BaseService):
    SERVICE_NAME = "github"
    API = "https://api.github.com"
    CLIENT_ID = '40435bdcc83a625b6c9b'
    CLIENT_SECRET = '81703a26b85a31127071c6bbbc54ac37ea970df3'

    def parent_branch_exists(self, token):
        path = '/repos/%s/%s/branches/%s' % (self.namespace, self.project, self.parent_branch)
        url = urljoin(str(self.API), str(path))
        try:
            res = requests.get(url)
        except Exception as e:
            self.logger.fatal(e)
            sys.exit()

        if res.status_code == 404:
            return False
        return True

    def issue_pull_request(self, obj):
        try:
            merge_message = obj['message']
            api_token = obj['api_token']
        except KeyError:
            self.logger.fatal("GitHubService::issue_pull_request has incorrect parameters @%s" % obj)
            sys.exit()

        params = {
            'title': merge_message,
            'head': self.source_branch,
            'base': self.parent_branch
        }

        path = 'repos/%s/%s/pulls' % (self.namespace, self.project)
        url = urljoin(str(self.API), str(path))
        headers = {"Authorization": "token %s" % api_token}
        res = {}

        try:
            res = requests.post(
                url=url,
                headers = headers,
                data = json.dumps(params)
                )
        except Exception as e:
            self.logger.fatal(e)
            sys.exit()

        json_response = json.loads(res.text)
        if res.status_code == 401:
            # Invalid credentials. Revoke token and start over.
            return (None, res.status_code)
        elif res.status_code >= 400:
            try:
                detailed_err_msg = json_response['errors'][0]['message']
            except:
                detailed_err_msg = ''

            try: 
                err_msg = json_response['message']
            except:
                err_msg = ''
            return ("Could not create merge request. %s. %s" % (err_msg, detailed_err_msg), res.status_code)

        elif res.status_code == 201:
            try:
                pr_href = json_response["_links"]["html"]["href"]
            except:
                pr_href = "Merge request succeeded, but no URL was returned."
            return (pr_href, None)
            
        else:
            return (json_response, None)


    def _req_user_pass(self):
        self.logger.warn("\n\n(One Time Setup) Please enter credentials to request API key")
        user = input("%s username: " % self.SERVICE_NAME)
        pw = getpass.getpass("%s password: " % self.SERVICE_NAME)
        return (user, pw)

    def _generate_fingerprint(self):
        # https://developer.github.com/changes/2014-12-08-removing-authorizations-token/
        # not a user-facing value, could be anything unique (per instance of repo)
        return random.random()

    def _setup_token(self):
        u, p = self._req_user_pass()
        fingerprint = self._generate_fingerprint()
        url = urljoin(self.API, 'authorizations/clients/%s/%s' % (self.CLIENT_ID, fingerprint))

        params = {
            'note': 'post-review cli utility (github, gitlab, etc)',
            'client_secret': self.CLIENT_SECRET,
            'scopes': ['public_repo', 'repo']
        }

        res = requests.put(
            url=url,
            auth = (u,p),
            data = json.dumps(params)
            )

        j = json.loads(res.text)
        if res.status_code >= 400:
            return (-1, j.get('message', 'UNDEFINED ERROR (no error description from server)'))
        else:
            try:
                if not j['token'] and j['hashed_token']:
                    #TODO: token is not in config, but its created. so its lost in the ether
                    return (-1, 'Token ID already in use or token lost in ether.\n'+
                        'Please re-try with a new Token ID or Please revoke all tokens available: https://github.com/settings/applications')
                else:
                    return (j['token'], None)
            except KeyError:
                return (-1, "Successful response but unable to fetch credentials.")



from __future__ import absolute_import
from builtins import object
from .logger import create_logger
from future.utils import with_metaclass

class Meta(type):
    def __str__(self):
        return self.SERVICE_NAME

class BaseService(with_metaclass(Meta, object)):
    SERVICE_NAME = None
    GIT_CONFIG_API_KEY = None
    API = None

    def __init__(self, source, parent, origin_domain, namespace, project, logger):
        self.source_branch = source
        self.parent_branch = parent
        self.origin_domain = origin_domain
        self.namespace = namespace
        self.project = project
        self.logger = logger

    

    def issue_pull_request(self, obj):
        raise NotImplementedError()

    def _setup_token(self):
        raise NotImplementedError()

    def parent_branch_exists(self, token):
        raise NotImplementedError()

    @classmethod
    def check_service_name(child_impl_class, service_name):
        return service_name == child_impl_class.SERVICE_NAME



from builtins import object
from .configprocesser import get_configuration
from .GitCommandRunner import GitCommandRunner as Git
from .GitServiceManager import GitServiceManager
from postreview import __version__

import argparse
import sys
import os

def main():
    driver = CliDriver()
    return driver.main()

class CliDriver(object):

    def main(self, args=None):

        if args is None:
            args = sys.argv[1:]

        parser = self._create_parser()
        parsed, remaining = parser.parse_known_args(args)
        try:
            if parsed.version:
                sys.stdout.write('post-review/' + __version__)
                sys.stdout.write('\n\n')
                return
            if parsed.target is None:
                raise ValueError()
            self.target = parsed.target
        except (AttributeError, ValueError):
            sys.stderr.write("===================================")
            sys.stderr.write("\n")
            sys.stderr.write("WARNING: missing --target argument")
            sys.stderr.write("\n")
            sys.stderr.write("===================================")
            sys.stderr.write("\n\n")
            parser.print_help()
            return 255

        git_service = GitServiceManager(self.target)
        return git_service.post_review()


    def _create_parser(self):
        parser = argparse.ArgumentParser(description="create a code review and merge request")
        #parser._action_groups.pop()
        required = parser.add_argument_group('Required Arguments')
        required.add_argument('--target', '-t', help='remote branch to diff/merge with.')
        parser.add_argument('--version', action='store_true', help='software package version installed')
        return parser



from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import input
from builtins import str
from .BaseService import BaseService

import json
import sys

#required
from urllib.parse import urljoin
from urllib.parse import quote
import requests


class GitLabService(BaseService):
    SERVICE_NAME = "gitlab"

    def _API(self):
        url = 'https://%s/' % self.origin_domain
        path = 'api/v4/'
        return urljoin(str(url), str(path))


    def parent_branch_exists(self, token):
        url_path = 'projects/%s/repository/branches/%s' % (self._url_encoded_path(), self.parent_branch)
        url = self._API() + url_path

        try:
            res = requests.get(
                url,
                headers={"PRIVATE-TOKEN": token}
            )
        except Exception as e:
            self.logger.fatal(e)
            sys.exit()

        if res.status_code == 404:
            return False
        return True


    def issue_pull_request(self, obj):
        try:
            merge_message = obj['message']
            api_token = obj['api_token']
        except KeyError:
            self.logger.fatal("GitLabService::issue_pull_request has incorrect parameters @%s" % obj)
            sys.exit()

        params = {
            'source_branch': self.source_branch,
            'target_branch': self.parent_branch,
            'title': merge_message
        }

        path = 'projects/%s/merge_requests' % self._url_encoded_path()
        url = urljoin(str(self._API()), str(path))
        headers = {"PRIVATE-TOKEN": api_token}
        res = {}

        try:
            res = requests.post(
                url,
                headers=headers,
                data=params
            )
        except Exception as e:
            self.logger.fatal(e)
            sys.exit()

        json_response = json.loads(res.text)
        if res.status_code == 401:
            # Invalid credentials. Revoke token and start over.
            return (None, res.status_code)
        elif res.status_code >= 400:
            try:
                err_msg = str(json_response['message'][0])
            except:
                err_msg = ''
            return ("Could not create merge request %s" % (err_msg), res.status_code)

        elif res.status_code == 201:
            try:
                pr_href = json_response["web_url"]
            except:
                pr_href = "Merge request succeeded, but no URL was returned."
            return (pr_href, None)

        else:
            return (json_response, None)


    def _url_encoded_path(self):
        # https://docs.gitlab.com/ce/api/README.html#namespaced-path-encoding
        return quote('%s/%s' % (self.namespace, self.project), safe='')


    def _setup_token(self):
        self.logger.warn("\n\n(One Time Setup) Please create a Personal Access Token")
        self.logger.warn("https://%s/profile/personal_access_tokens" % self.origin_domain)
        self.logger.warn("Scope: API, Expires: Never\n")
        token = input("Please enter your Personal Access Token:  ")

        # Make request to resource that requires us to be authenticated
        path = 'projects/%s/labels' % self._url_encoded_path()
        url = urljoin(str(self._API()), path)

        res = requests.get(
            url,
            headers={"PRIVATE-TOKEN": token}
            )

        if res.status_code == 200:
            return(token, None)
        return(-1, "Invalid Personal Access Token")



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from configparser import SafeConfigParser, NoSectionError, NoOptionError
import sys
from .logger import create_logger
import os

#Get path relevant to where configprocesser.py is
def _config_path():
    mydir = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(mydir, 'config.ini')

#TODO: Throw errors if not configuration or settings
def _fetch(section, key):
    config = SafeConfigParser()
    config.read(_config_path())

    try:
        data = config.get(section, key)
        return data
    except NoSectionError:
        #Logging Error
        logger.fatal("%s section does not exist in config.ini" % section)
        sys.exit()
    except NoOptionError:
        #TODO LOGGING ERROR
        logger.fatal("%s key does not exist in config.ini" % key)
        sys.exit()

#def _insert(section, key, value):
#    config = SafeConfigParser()
#    config.read(_config_path())
#    config.set(section, key, value)

#    with open(_config_path(), 'wb') as configfile:
#        config.write(configfile)
#    return 1

#def put_config(section, key, value):
#    return _insert(section, key, value)

def get_configuration(arg):
    return _fetch('configuration', arg)


#!/usr/bin/env python

from future import standard_library
standard_library.install_aliases()

import sys
import os

if os.environ.get('LC_CTYPE', '') == 'UTF-8':
    os.environ['LC_CTYPE'] = 'en_US.UTF-8'
import cli

def main():
    return cli.main()

if __name__ == '__main__':
    sys.exit(main())


#!/usr/bin/env python

from setuptools import setup, find_packages
import re
import codecs
import os.path
import sys

requires = ['requests==2.18.1',
            'future==0.16.0',
            'coloredlogs==7.1',
            'configparser==3.5.0'
            ]


here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


#Unable to use just Manifest because config.ini exists in a package
#Must use both package_data and Manifest

setup_options = dict(
    name='post-review',
    version=find_version("postreview", "__init__.py"),
    description='unified command line interface for posting code reviews and merge requests',
    long_description=open('README.rst').read(),
    url='https://github.com/ericforbes/post-review',
    author='Eric Forbes',
    author_email='ericforbes91@gmail.com',
    license='MIT',
    keywords='git devops review',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    package_data={
        'postreview': ['config.ini']
    },
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,<4',
    entry_points={
        'console_scripts': [
            'post-review=postreview.__main__:main'
        ]
    },
    install_requires=[
        'requests>=2.18',
        'coloredlogs>=7.1',
        'configparser>=3.5.0',
        'future>=0.16.0'
        ],
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Version Control :: Git',
        )
    )

setup(**setup_options)


import coloredlogs
import logging
import sys

coloredlogs.install(fmt='%(message)s')

def create_logger():
    #logging.basicConfig(format='%(levelname)s - %(message)s')
    logging.basicConfig(format='%(message)s')

    root = logging.getLogger()
    root.setLevel(logging.getLevelName('INFO'))

    #Add handler for standard output (console) any debug+
    #ch = logging.StreamHandler(sys.stdout)
    #ch.setLevel(logging.getLevelName('DEBUG'))
    #formatter = logging.Formatter('%(message)s')
    #ch.setFormatter(formatter)
    #handler = ColorStreamHandler()
    #handler.setLevel(logging.getLevelName("DEBUG"))
    #root.addHandler(handler)
    
    return root



from future import standard_library
standard_library.install_aliases()
import re
from urllib.parse import urlparse

def determine_git_domain(origin_url):
        #Handles both SSH and HTTPS
        #No support for relative URL path, only (sub)domain: https://docs.gitlab.com/omnibus/settings/configuration.html

        #HTTPS URLS
        url = urlparse(origin_url)
        if url.netloc:
            return url.netloc

        #SSH URLS
        matchObj = re.search(r'@+(?P<ssh_git_url>.*)?:', origin_url)

        try:
            if matchObj.group('ssh_git_url'):
                return matchObj.group('ssh_git_url')
        except:
            return None


def repo_url_attributes(domain, remote_origin_url):
        domain_fixed = re.escape(domain)

        # git@gitlab.com:ericforbes/my-postreview-test.git -> ericforbes/
        # https://gitlab.com/ericforbes/private/my-postreview-test.git -> ericforbes/private/
        namespace_match = re.search(r'%s(:|\/)(?P<namespace>.*\/)' % domain_fixed, remote_origin_url)

        # git@gitlab.com:ericforbes/private/my-postreview-test.git -> my-postreview-test
        # https://gitlab.com/ericforbes/my-postreview-test.git -> my-postreview-test
        project_match = re.search(r'%s(:|\/).*\/(?P<project>.*).git' % domain_fixed, remote_origin_url)

        try:
            # Remove any leading or trailing slashes -> ericforbes/private
            namespace = namespace_match.group('namespace').strip("/")

            project = project_match.group('project')

            return (namespace, project)
        except:
            return (None, None)
